namespace NXyn.EmitFSharp
  open System
  open System.Reflection
  open System.Collections
  open System.Collections.Generic

  open NXyn
  open NXyn.EmitFSharp.Common
  open NXyn.Aux.IOMisc

  module JSONio =
    (* suitable parsers and unparsers in JSON module; non-portable builtin types like nativeint are not supported *)
    let builtinFSharpTypes = [| typeof<bigint>,   "Value.BigInt",  "BigInt_of_Value";
                                typeof<bool>,     "Value.Bool",    "Bool_of_Value";
                                typeof<byte>,     "Value.Byte",    "Byte_of_Value";
                                typeof<char>,     "Value.Char",    "Char_of_Value";
                                typeof<decimal>,  "Value.Decimal", "Decimal_of_Value";
                                typeof<double>,   "Value.Double",  "Double_of_Value";
                                typeof<int16>,    "Value.Int16",   "Int16_of_Value";
                                typeof<int32>,    "Value.Int32",   "Int32_of_Value";
                                typeof<int64>,    "Value.Int64",   "Int64_of_Value";
                                typeof<sbyte>,    "Value.SByte",   "SByte_of_Value";
                                typeof<single>,   "Value.Single",  "Single_of_Value";
                                typeof<string>,   "Value.String0", "String0_of_Value";     (* also allow null strings *)
                                typeof<uint16>,   "Value.UInt16",  "UInt16_of_Value";
                                typeof<uint32>,   "Value.UInt32",  "UInt32_of_Value";
                                typeof<uint64>,   "Value.UInt64",  "UInt64_of_Value"; |]

    let ModulePrelude = String.Join(Environment.NewLine, [|"open NXyn.JSON"; "open System.Collections.Generic" |])
(* TBD should not make any assumptions about opened modules !!! *)

    let userParserName = "FromJSON"
    let userUnparserName = "ToJSON"

    (***)

    let hasPublicStaticMethod t (name:string) (args:Type []) ret = let m = (t:Type).GetMethod(name,args)
                                                                   m <> null &&
                                                                   ret = m.ReturnType &&
                                                                   let staticPublicAtts = MethodAttributes.Static ||| MethodAttributes.Public
                                                                   staticPublicAtts = (m.Attributes &&& staticPublicAtts)

    let initMap t selBuiltin userDef check = let (userDefined0,need) = CollectTypes2(t) |> Array.filter (isBuiltinFSharpType >> not) |> Array.partition check
                                             let needed      = need               |> Array.mapi (fun i t     -> (t, sprintf "M%d" i))
                                             let builtin     = builtinFSharpTypes |> Array.map  (fun (t,u,p) -> (t, selBuiltin(u,p)))
                                             let userDefined = userDefined0       |> Array.map  (fun t       -> (t, sprintf "%s.%s" (fsharpTypeName_of_Type t) userDef))
                                             let map = dict(Array.concat([builtin; needed; userDefined]))
                                             (map,needed)

    let unsupported where (t:Type) = if   t.IsByRef   then failwithf "EmitJSONio.%s: byref type '%A' not supported"   where t
                                     elif t.IsPointer then failwithf "EmitJSONio.%s: pointer type '%A' not supported" where t
                                     elif t.IsClass   then failwithf "EmitJSONio.%s: class type '%A' not supported"   where t
                                                      else failwithf "EmitJSONio.%s: type '%A' not supported"         where t
    (***)

    let generateCodeAndEmit w pre emit1 todo post = (w:IO.TextWriter).Write(pre:string)
                                                    todo
                                                    |> Array.map (fun el -> async { return(let w = new IO.StringWriter() in emit1 w el; w.ToString()) })
                                                    |> Async.Parallel
                                                    |> Async.RunSynchronously
                                                    |> Array.iteri (fun i s -> fprintf w "  %s %s" (if i = 0 then "let rec" else "and") s)
                                                    w.Write(post:string)
    (***)

    let WriteUnparser (name:string) (t:Type) (w:IO.TextWriter) =
        let hasUserDefinedUnparser t = hasPublicStaticMethod t userUnparserName [|t|] typeof<JSON.Value>
        let (map,todo) = initMap t fst userUnparserName hasUserDefinedUnparser
        let emit1 w (t,map_t) =
            fprintf w "%s" map_t
            (* order is important here *)
            if   t = typeof<BitArray> then fprintfn w "(x) = (* BitArray *) Value.BitArray(x)"
            elif t.IsEnum             then fprintfn w "(x) = (* Enum *) Value.String(System.Enum.Format(typeof<%s>, x, \"f\"))" (fsharpTypeName_of_Type(t))
            elif t.IsArray            then fprintfn w "(x) = (* Array *) Value.Array(Array.map %s x)"                           map.[elementType_of_Type(t)]
            elif is_FSharpList(t)     then fprintfn w "(x) = (* F# list *) Value.Array(Array.map %s (Array.ofList(x)))"         map.[firstGenericArgument_of_Type(t)]
            elif is_FSharpOption(t)   then fprintfn w "(x) = (* F# option *) Value.Array(Array.map %s (Option.toArray(x)))"     map.[firstGenericArgument_of_Type(t)]
            elif is_FSharpTuple(t)    then let writeElements_to = WriteManyI_to "" ";" "" (fun i t w -> fprintf w "%s(a%d)" map.[t] i)
                                           let tels = tupleElements_of_Type(t)
                                           fprintfn w "((%s):%s) = (* F# tuple *) Value.Array([|%t|])"
                                                      (tels.Length |> Aux.IOMisc.string_aI "" "")
                                                      (t |> fsharpTypeName_of_Type)
                                                      (tels |> writeElements_to) (* WriteManyI_to "" ";" "" (fun i t w -> fprintf w "%s(a%d)" map.[t] i)) *)
            elif is_FSharpUnion(t)    then fprintfn w " = function (* F# union *)"
                                           let tName = fsharpTypeName_of_Type(t)
                                           let ucs = unionCases_of_Type(t)
                                           for uc in ucs do let ps = propertyType_of_UnionCaseInfo uc
                                                            fprintfn w "    | %s.%s%s -> %sValue.String(\"%s\")%t"
                                                                       tName
                                                                       uc.Name
                                                                       (ps.Length |> Aux.IOMisc.string_aI "(" ")")
                                                                       (if ps = [||] then "" else "Value.Array([|")
                                                                       uc.Name
                                                                       (ps |> WriteManyI_to ";" ";" "|])" (fun i t w -> fprintf w "%s(a%d)" map.[t] i))
            elif is_ICollection(t)    then let collType = firstGenericArgument_of_Type(interface_of_Type "ICollection`1" t)
                                           fprintfn w "(x:%s) = (* ICollection *) let xs = Array.zeroCreate x.Count in (x :> ICollection<%s>).CopyTo(xs,0); Value.MappedArray(%s,xs)"
                                                      (t |> fsharpTypeName_of_Type)
                                                      (collType |> fsharpTypeName_of_Type)
                                                      map.[collType]
            elif is_IEnumerable(t)    then fprintfn w "(x) = (* IEnumerable *) Value.Array(Seq.toArray (Seq.map %s x))"
                                                      map.[firstGenericArgument_of_Type (interface_of_Type "IEnumerable`1" t)]
            elif is_FSharpRecord(t)   then fprintfn w "(x:%s) = (* F# record *) Value.Assoc([|%t|])"
                                                      (fsharpTypeName_of_Type(t))
                                                      (propertyMembers_of_Type(t) |> WriteMany_to "" ";" "" (fun (m:MemberInfo) w -> fprintf w "\"%s\",%s(x.%s)" m.Name map.[t.GetProperty(m.Name).PropertyType] m.Name))
            elif is_Struct(t)         then let fprintMembers = WriteMany_to "" ";" "" (fun (m:MemberInfo) w -> fprintf w "%s(x.%s)" map.[t.GetProperty(m.Name).PropertyType] m.Name)
                                           fprintfn w "(x:%s) = (* struct *) Value.Array([|%t|])"
                                                      (fsharpTypeName_of_Type(t))
                                                      (propertyMembers_of_Type(t) |> fprintMembers)
                                      else unsupported "emitUnparser" t
        generateCodeAndEmit w (sprintf "let %s(z:%s) : Value =\n" name (fsharpTypeName_of_Type(t))) emit1 todo (sprintf "  z |> %s\n;;\n\n" map.[t])

    (***)

    let WriteParser (name:string) (t:Type) (w:IO.TextWriter) =
        let hasUserDefinedParser t = hasPublicStaticMethod t userParserName [|typeof<JSON.Value>|] t
        let (map,todo) = initMap t snd userParserName hasUserDefinedParser
        let emit1 w (t,map_t) =
            let tName = fsharpTypeName_of_Type'(t) (* TOO MUCH?, module not needed *)
            fprintf w "%s" map_t
            (* order is important here *)
            if   t = typeof<BitArray> then fprintfn w "(x) = (* BitArray *) BitArray_of_Value(x)"
            elif t.IsEnum             then fprintfn w "(x) = (* Enum *) System.Enum.Parse(typeof<%s>, String_of_Value(x)) :?> %s"    tName tName
            elif t.IsArray            then fprintfn w "(x) = (* Array *) MappedArray_of_Value %s x"                                  map.[elementType_of_Type(t)]
            elif is_FSharpList(t)     then fprintfn w "(x) = (* F# list *) Array.toList(MappedArray_of_Value %s x)"                  map.[firstGenericArgument_of_Type(t)]
            elif is_FSharpOption(t)   then fprintfn w "(x:Value) = (* F# option *) if x.Length < 1 then None else Some(x.[0] |> %s)" map.[firstGenericArgument_of_Type(t)]
            elif is_FSharpTuple(t)    then let ts = tupleElements_of_Type(t)
                                           fprintfn w "(x:Value) = (* F# tuple *) if x.Length = %d then %t else failwith \"parse error tuple %s: wrong arity\""
                                                      ts.Length
                                                      (ts |> WriteManyI_to "(" "," ")" (fun i t w -> fprintf w "%s(x.[%d])" map.[t] i))
                                                      tName
            elif is_FSharpUnion(t)    then let ucInit = new IO.StringWriter()
                                           let ucs = unionCases_of_Type(t)
                                           fprintfn w "_UCdict = new Dictionary<string,int * (Value -> %s)>(%d)" tName ucs.Length
                                           let writeMapUC = WriteManyI_to "(" "," ")" (fun i t w -> fprintf w "%s(x.[%d])" map.[t] (i+1))
                                           for uc in ucs do let ts = propertyType_of_UnionCaseInfo uc
                                                            fprintfn ucInit "    %s_UCdict.Add(\"%s\", (%d,%s_UC_%s))" map_t uc.Name ts.Length map_t uc.Name
                                                            fprintfn w "  and %s_UC_%s(x:Value) = %s.%s%t" map_t uc.Name tName uc.Name (ts |> writeMapUC)
                                           fprintf  w "  and %s_UCinit =\n%s" map_t (ucInit.ToString())
                                           fprintfn w "  and %s(x:Value) = (* F# union *)" map_t
                                           fprintfn w "    let ucName = String_of_Value(if x.IsString then x else x.[0])"
                                           fprintfn w "    let (n,f) = %s_UCdict.[ucName]" map_t
                                           fprintfn w "    if (x.IsString && n = 0) || x.Length = n+1 then f(x) else failwithf \"parse error %s.%%s: wrong arity: got %%d expected %%d (%%A)\" ucName x.Length (n+1) x" tName

            elif is_ICollection(t)    then fprintfn w "(x:Value) : %s = (* System.Collections.Generic.ICollection *)" tName
                                           let hasEmptyConstructor = t.GetConstructors() |> Array.exists (fun c -> [||] = c.GetParameters())
                                           let enumType = firstGenericArgument_of_Type(interface_of_Type "ICollection`1" t)
                                           fprintfn w "    let collection = new %s(%s)" tName (if hasEmptyConstructor then "" else "[||]")
                                           fprintfn w "    let icollection = collection :> ICollection<%s>" (enumType |> fsharpTypeName_of_Type)
                                           fprintfn w "    for i = 0 to x.Length - 1 do icollection.Add(%s(x.[i]))" map.[enumType]
                                           fprintfn w "    collection"
            elif is_FSharpRecord(t)   then fprintfn w "(x:Value) = (* F# record *) assert(x.IsAssoc); %t"
                                                      (propertyMembers_of_Type(t) |> WriteMany_to "{" ";" "}" (fun m w -> fprintf w "%s.%s = %s(x.[\"%s\"])" ((* BOILER? *) ClippedFSharpTypeName(tName)) m.Name map.[t.GetProperty(m.Name).PropertyType] m.Name))
            elif is_Struct(t)         then let ts = propertyMembers_of_Type(t) |> Array.map (fun (m:MemberInfo) -> t.GetProperty(m.Name).PropertyType)
                                           if ts <> [||] then if null = t.GetConstructor(ts) then failwithf "emitP_struct: constructor missing for struct-type %A" t
                                                              if Seq.length(Seq.distinct(ts)) <> ts.Length then NXyn.Aux.IOMisc.StdErr.WriteLine(sprintf "emitP_struct: ambiguous constructor for %A" t)
                                           fprintfn w "(x:Value) = (* struct *) if x.Length = %d then new %s(%t) else failwith \"parse error struct %s: wrong arity\""
                                                      ts.Length
                                                      tName
                                                      (ts |> WriteManyI_to "" "," "" (fun i t w -> fprintfn w "%s(x.[%d])" map.[t] i))
                                                      tName
                                      else unsupported "emitParser" t
        generateCodeAndEmit w (sprintf "let %s(z:Value) : %s =\n" name (fsharpTypeName_of_Type(t))) emit1 todo (sprintf "  z |> %s\n;;\n\n" map.[t])
(* END *)
