module Genes01

open System

let cutLast (nbits:int) (d:int64 array) =
  if nbits &&& 63 = 0 then d 
  else
    let last_ix = ((nbits + 63) >>> 6) - 1
    d.[last_ix] <- d.[last_ix] &&& ((1L <<< (nbits &&& 63)) - 1L)
    d;;

let makeBits (nbits:int) (w:int64) = cutLast nbits (Array.create ((nbits + 63) >>> 6) w);;

let getBit   (s:int64 array) (i:int) = s.[i >>> 6] &&& (1L <<< (i &&& 63)) <> 0L;;

let flipBit  (d:int64 array) (i:int) = d.[i >>> 6] <- d.[i >>> 6] ^^^ (1L <<< (i &&& 63));;
let setBit   (d:int64 array) (i:int) = d.[i >>> 6] <- d.[i >>> 6] ||| (1L <<< (i &&& 63));;
let clearBit (d:int64 array) (i:int) = d.[i >>> 6] <- d.[i >>> 6] &&& (~~~ (1L <<< (i &&& 63)));;

let setOrClearBit (v:bool) (d:int64 array) (i:int) = if v then setBit d i else clearBit d i;;

let popCount (x:int64) =
  let x = x - ((x >>> 1) &&& 0x5555555555555555L)
  let x = (x &&& 0x3333333333333333L) + ((x >>> 2) &&& 0x3333333333333333L)
  let x = (x + (x >>> 4)) &&& 0x0f0f0f0f0f0f0f0fL
  int ((x * 0x0101010101010101L) >>> 56);;

let sumHammingDistance (sum0:int) (x:int64) (y:int64) = sum0 + popCount (~~~ (x ^^^ y));;

let bitRev (x:int64) =
  let x = ((x >>> 1)  &&& 0x5555555555555555L) ||| ((x &&& 0x5555555555555555L) <<< 1)
  let x = ((x >>> 2)  &&& 0x3333333333333333L) ||| ((x &&& 0x3333333333333333L) <<< 2)
  let x = ((x >>> 4)  &&& 0x0F0F0F0F0F0F0F0FL) ||| ((x &&& 0x0F0F0F0F0F0F0F0FL) <<< 4)
  let x = ((x >>> 8)  &&& 0x00FF00FF00FF00FFL) ||| ((x &&& 0x00FF00FF00FF00FFL) <<< 8)
  let x = ((x >>> 16) &&& 0x0000FFFF0000FFFFL) ||| ((x &&& 0x0000FFFF0000FFFFL) <<< 16)
  ((x >>> 32) &&& 0x00000000FFFFFFFFL) ||| ((x &&& 0x00000000FFFFFFFFL) <<< 32);;

let makeRandomBits (r:Random) (nbits:int) =
  let d = makeBits nbits 0L
  let tmp = Array.create 8 0uy
  for i = 0 to d.Length - 1 do
    r.NextBytes(tmp)
    d.[i] <- System.BitConverter.ToInt64(tmp,0)
  cutLast nbits d;;

type t(s:int64 array,nbits:int) =
  new (str:string) =
    let d = makeBits str.Length 0L
    for i = 0 to str.Length - 1 do
      match str.[i] with
        | '0' -> ()
        | '1' -> setBit d i
        | _   -> failwith "Genes01.new: invalid char"
    t(d, str.Length)

  member x.Item with get (ix:int) = getBit s ix

  member x.MutateSet1 (ix:int)   = setBit s ix
  member x.MutateFlip1 (ix:int)  = flipBit s ix
  member x.MutateClear1 (ix:int) = clearBit s ix

  member x.MutateSetN (ixFrom:int, ixTo:int) =
    let (wA,wZ) = (ixFrom >>> 6, ixTo >>> 6)
    if wA = wZ then s.[wZ] <- s.[wZ] ||| ((1L <<< ((ixTo &&& 63)+1)) - (1L <<< ixFrom))
    else
      s.[wA] <- s.[wA] ||| (~~~ ((1L <<< ixFrom) - 1L))
      for i = wA + 1 to wZ - 1 do s.[i] <- -1L
      s.[wZ] <- s.[wZ] ||| ((1L <<< ixTo) - 1L)

  member x.MutateClearN (ixFrom:int, ixTo:int) =
    let (wA,wZ) = (ixFrom >>> 6, ixTo >>> 6)
    if wA = wZ then s.[wZ] <- s.[wZ] &&& (~~~ ((1L <<< ((ixTo &&& 63)+1)) - (1L <<< ixFrom)))
    else
      s.[wA] <- s.[wA] &&& ((1L <<< ixFrom) - 1L)
      for i = wA + 1 to wZ - 1 do s.[i] <- 0L
      s.[wZ] <- s.[wZ] &&& (~~~ ((1L <<< ixTo) - 1L))

  member x.MutateFlipN (ixFrom:int, ixTo:int) =
    let (wA,wZ) = (ixFrom >>> 6, ixTo >>> 6)
    if wA = wZ then s.[wZ] <- s.[wZ] ^^^ ((1L <<< ((ixTo &&& 63)+1)) - (1L <<< ixFrom))
    else
      s.[wA] <- s.[wA] ^^^ (~~~ ((1L <<< ixFrom) - 1L))
      for i = wA + 1 to wZ - 1 do s.[i] <- ~~~ s.[i]
      s.[wZ] <- s.[wZ] ^^^ ((1L <<< ixTo) - 1L)

  member x.MutateReverseN (ixFrom:int, ixTo:int) = (* SLOW *)
    for i = 0 to (ixTo - ixFrom - 1) / 2 do
      let t = x.[ixFrom+i]
      setOrClearBit x.[ixTo-i] s (ixFrom+i) 
      setOrClearBit t s (ixTo-i) (* ??? *)

  member x.Size = nbits
  member private x.Data = s

  member x.Count = Array.sumBy popCount s
  member x.HammingDistance (y:t) = Array.fold2 sumHammingDistance 0 x.Data y.Data

  static member Random (rng:Random,nbits:int) = t(makeRandomBits rng nbits,nbits)
  static member Replicate (w:int64, nbits:int) = t(makeBits nbits w, nbits)

  member x.CrossoverUniform (y:t,sel:t) = (* bitwise crossover as determined by `sel' *)
    let d1 = makeBits nbits 0L
    let d2 = makeBits nbits 0L
    for i = 0 to x.Data.Length - 1 do 
      d1.[i] <- (x.Data.[i] &&& sel.Data.[i]) ||| (y.Data.[i] &&& (~~~ sel.Data.[i]))
      d2.[i] <- (x.Data.[i] &&& (~~~ sel.Data.[i])) ||| (y.Data.[i] &&& sel.Data.[i])
    (t(d1,nbits), t(d2,nbits))

  member x.CrossoverMulti (positions:int [], y:t) = (* n-point crossover *)
    let d1 = x.Data.Clone() :?> int64 []
    let d2 = y.Data.Clone() :?> int64 []
    positions
    |> Array.iter (fun(pos) ->
                     let (w,b) = (pos >>> 6, pos &&& 63)
                     let (x_w,y_w) = (d1.[w], d2.[w])
                     let m = (1L <<< b) - 1L
                     d1.[w] <- (x_w &&& m) ||| (y_w &&& (~~~ m))
                     d2.[w] <- (y_w &&& m) ||| (x_w &&& (~~~ m))
                     for i = w + 1 to d1.Length - 1 do
                           let t = d1.[i]
                           d1.[i] <- d2.[i]
                           d2.[i] <- t)
    (t(d1,nbits), t(d2,y.Size))

  interface System.ICloneable with
      member x.Clone () = t((s.Clone () :?> int64 array), nbits) :> obj

  member x.DeepCopy() = t((s.Clone () :?> int64 array), nbits)
      

  override x.Equals (y:obj) = 
    y <> null && 
    (x.GetType() = y.GetType()) && 
    (let y = y :?> t in x.Size = y.Size && x.Data = y.Data)

  override x.GetHashCode () = (* Austin Appleby, MurmurHash2 *)
    (* 'm' and 'r' are mixing constants generated offline. They're not really 'magic', they just happen to work well. *)
    let m = 0x5bd1e995u
    let r = 24

    let mutable h = uint32 s.Length (* Initialize the hash to a 'random' value *)

    for i = 0 to s.Length - 1 do  (* Mix 4 bytes at a time into the hash *)
      let k = s.[i]
      let (kL,kH) = (uint32 k, uint32 (k >>> 32))
      let kL = kL * m
      let kL = kL ^^^ (kL >>> r)
      let kH = kH * m
      let kH = kH ^^^ (kH >>> r)
      h <- (h * m) ^^^ (kL * m)
      h <- (h * m) ^^^ (kH * m)
    int h



  override x.ToString () =
    let mutable allequal = true     (* try to detect replicate *)
    let mutable refelem = s.[0]
    let mutable i = 1
    while allequal && i < s.Length - 1 do
      allequal <- s.[i] = refelem
      i <- i + 1
    if allequal && s.[s.Length-1] = (refelem &&& ((1L <<< (nbits &&& 63)) - 1L)) then
      sprintf "Genes01.t.Replicate(0x%xL,%d)" refelem nbits
    else
      let mutable buf = new System.Text.StringBuilder(nbits)
      for i = 0 to nbits - 1 do
        buf <- buf.Append (if x.[i] then '1' else '0')
      sprintf "Genes01.t(\"%s\")" (buf.ToString ())
;;

