module OptSimple

open System
open System.Collections.Generic

open Ast
open X64instr

open InstrIO
open MoreInstrIO

open AstIO
open MoreAstIO


open NXyn
open NXyn.Aux
open NXyn.Aux.Global

open NXyn.Aux.IntMisc
open NXyn.Aux.OtherMisc
open NXyn.Aux.SetMisc

open Basics

open LocalScheduling


let elimDC_returnLiveness (keepStores:bool) (tDict:Dictionary<String,IBlock>) (cc:String []) =
    (* propagate liveness information *)
    let inoutDict = new Dictionary<String,Set<Operand> * Set<Operand>>()
    let get_IN_of_label(l) = if   externABI.Contains(l)    then begin eprintf "abi %s" l; ABIParamRegistersSet end
                             elif inoutDict.ContainsKey(l) then inoutDict.[l] |> fst
                                                           else Set.empty
    let instr_is_live keep out s ds d i = keep ||
                                          not(Set.isEmpty(Set.intersect out ds)) || not(Set.isEmpty(Set.intersect out d)) ||
                                          Instr_is_Prefix(i) || Instr_uses_FixedMem(i) || Instr_may_raise_exn(i) || Instr_is_Jfollow(i) || Instr_is_Jthru(i) ||
                                          (if keepStores then Md_of_Instr(i) <> [] || Mds_of_Instr(i) <> []
                                           else
                                                Md_of_Instr(i)  |> List.exists (stackLocOption >> Option.isNone) ||
                                                Mds_of_Instr(i) |> List.exists (stackLocOption >> Option.isNone)
                                           ) ||
                                          Set.contains Operand.rsp ds || Set.contains Operand.rsp d
    let prop1 keep b i (out,instrs) = let s  = S_of_Instr  konstSizeDict b.ReturnsVoid i
                                      let ds = DS_of_Instr i
                                      let d  = D_of_Instr  i
                                      if instr_is_live keep out s ds d i then (Set.union (Set.union s ds) (Set.difference out d), i::instrs)
                                                                         else // LogInfo(sprintf "dropping instruction %A" i)
                                                                              (out,instrs)
    let propagate_liveness keep changed l = let b = tDict.[l]
                                            (b.DynSuccs,Set.empty)
                                            ||> Array.foldBack (fun l -> Set.union (get_IN_of_label(l)))
                                            |> fun out1 -> (b.Instrs,(out1,[]))
                                                           ||> Array.foldBack (prop1 keep b)
                                                           |> fun (in1,instrs) -> let (found,inout0) = inoutDict.TryGetValue(l)
                                                                                  let inout1 = (in1,out1)
                                                                                  if keep then inoutDict.[l] <- inout1
                                                                                          else b.Instrs  <- Array.ofList(instrs)
                                                                                  changed || not(found) || inout0 <> inout1
    (* iterate until fixed point is reached *)
    while (false,cc) ||> Array.fold (propagate_liveness true) do ()

    (* final pass for eliminating useless instructions *)
    (false,cc) ||> Array.fold (propagate_liveness false) |> ignore

    // dump info
    (*
    eprintfn "----------------------------------------------------------------------------------------"
    cc |> Array.iteri (fun i xs -> eprintfn "\ncomp #%d: %A" i xs
                                   for lbl in xs do
                                       if not(tDict.ContainsKey(lbl)) then eprintfn "output: not connected %s" lbl
                                       else if not(inoutDict.ContainsKey(lbl)) then eprintfn "output: not inoutDict %s" lbl
                                       else eprintfn "lbl: %s" lbl
                                            let block = tDict.[lbl]
                                            let instrs = block.Instrs
                                            if instrs.Length > 0 then let lastinstr = instrs.[instrs.Length - 1]
                                                                      eprintfn "\tninstr     : %d" instrs.Length
                                                                      eprintfn "\tlast-instr : %A" lastinstr
                                            eprintfn "\tretvoid    : %A" block.ReturnsVoid
                                            eprintf  "\tdynpreds   : ["; for x in block.DynPreds do eprintf "%A " x done; eprintfn "]"
                                            eprintf  "\tdynsuccs   : ["; for x in block.DynSuccs do eprintf "%A " x done; eprintfn "]"
                                            eprintf  "\tin         : [ "; for x in fst(inoutDict.[lbl]) do eprintf "%s " (x.ToString()) done; eprintfn "]"
                                            eprintf  "\tout        : [ "; for x in snd(inoutDict.[lbl]) do eprintf "%s " (x.ToString()) done; eprintfn "]")
*)
    (* return dictionary of all block inputs/outputs *)
    inoutDict





let nextReader_of' x instrs =
                   let rec loop bag = function [] -> None
                                             | (instr0,aux0,s0,ds0,d0)::instrs -> if   Set.contains x s0  then Some(bag,(instr0,aux0,s0,ds0,d0),instrs)
                                                                                  elif Set.contains x ds0 then Some(bag,(instr0,aux0,s0,ds0,d0),instrs)
                                                                                  elif Set.contains x d0  then None
                                                                                                          else loop ((instr0,aux0,s0,ds0,d0)::bag) instrs
                   loop [] instrs

let nextReader_of (b:IBlock) x instrs =
                  let rec loop bag = function [] -> None
                                            | i::instrs -> if   Set.contains x (S_of_Instr  konstSizeDict b.ReturnsVoid i) then Some(bag,i,instrs)
                                                           elif Set.contains x (DS_of_Instr i) then Some(bag,i,instrs)
                                                           elif Set.contains x (D_of_Instr  i) then None
                                                                                               else loop (i::bag) instrs
                  loop [] instrs

(***)

let maxN = if TargetIsAMD() then 2 else 1 (* konstant, TBD good name *)


// name?
type Raccess = LocalRO | LocalRW of (int * int) | NonLocal



// -> X64instr
let Raccess_of_M (size,m) =
                 match m
                  with | K(Id({Name = k})) | K64rip(Id({Name = k})) when konstSizeDict.ContainsKey(k) -> LocalRO
                       | m -> match stackLocOption(size,m)
                               with | Some(m) -> LocalRW(m)
                                    | None -> NonLocal

let rec usedX_beforeX_in d s = function
    | [] -> false
    | i::instrs ->
        let x_in_i = X_of_Instr(i)
        if   List.exists ((=)d) x_in_i then true
        elif List.exists ((=)s) x_in_i then false
                                       else usedX_beforeX_in d s instrs



let swap_copy_operands (in0,out0) (instrs:Instr []) = // tDict:Dictionary<String,IBlock>) (inoutDict:Dictionary<String,Set<Operand> * Set<Operand>>) =
        let rec loop bag = function
            | [] -> List.rev(bag)
            | i::instrs ->
                match i with
                    | XsXd(((I_XsXd.Movaps | I_XsXd.Movapd) as op1), {Iargs_S_D.S=s1;Iargs_S_D.D=d1})
                      when s1 <> d1
//                        && not (Set.contains (Operand_of_X s1) in0) // TBD assert that no fixedX!!!
//                        && not (Set.contains (Operand_of_X d1) in0)
                        && not (Set.contains (Operand_of_X s1) out0)
                        && not (Set.contains (Operand_of_X d1) out0)
                        && usedX_beforeX_in d1 s1 instrs
                           -> eprintf "/"
                              loop (i::bag) (List.map (Map_X_in_Instr (swap2 s1 d1)) instrs)
                    | _ -> loop (i::bag) instrs

        let instrs = loop [] (Array.toList instrs)
        Array.ofList(instrs)
    

let elim_useless_vec_copies_block ((in0,out0):Set<Operand> * Set<Operand>) (b:IBlock) =
        if Array.exists (function XsXd _ -> true | _ -> false) b.Instrs then
           let out0 = Set.filter (function Vector(v) -> true | _ -> false) out0 (* vector only *)
           let in0  = Set.filter (function Vector(v) -> true | _ -> false) in0  (* vector only *)

           Set.empty
           |> Array.foldBack Set.union (Array.map (fun(i) -> D_of_Instr i) b.Instrs) 
           |> fun written -> Set.difference written out0
           |> fun written ->
                let rec loop bag = function
                    | [] -> List.rev bag
                    | i::instrs ->
                        match i with
                            (* bUGGY, wrong place to do this ? (better w/AI info!) *)
                            | M128sXd((I_M128sXd.Movaps | I_M128sXd.Movapd (* ??? *)), {Iargs_S_D.S=M128.S(m0) as s0; Iargs_S_D.D=d0})
                              when not(Set.contains (Operand_of_X d0) out0)
//                              when false && not(Set.contains (Operand_of_X d0) out0)
//                                && NonLocal <> Raccess_of_M(16,m0)

                              ->
                                let access = Raccess_of_M(16,m0)
//DEL
//                                let access = match s0 with M128.S(K(Id({Name = k})) | K64rip(Id({Name = k}))) when konstSizeDict.ContainsKey(k) -> LocalRO
//                                                         | M128.S(m) -> match stackLocOption(16,m) with
//                                                                            | Some(m) -> LocalRW(m)
//                                                                            | None -> NonLocal


                                (* inline_vec_loads *)
//                                let rec limitedmap n bag instrs4 = if n > maxN then None
//                                                                               else eprintfn "limitedmap %d %A %d" n bag (List.length instrs4)
//                                                                                    limitedmap2 n bag instrs4
                                let rec limitedmap n bag instrs =
                                          if n > maxN then None
                                          else
                                              //eprintfn "limitedmap %d %A %d" n bag (List.length instrs)
                                              match instrs with [] -> Some(bag, [])
                                              
                                                               | (XsXds(op1, {Iargs_S_DS.S=s1; Iargs_S_DS.DS=ds1}))::instrs
                                                                 when s1 = d0
                                                                 (* + swap if commutative *)
                                                                      -> if s1 = ds1 then None
                                                                                     else match try__M128sXds_of_XsXds(op1)
                                                                                           with | None -> None
                                                                                                | Some(op1') -> let i1' = (M128sXds(op1',{Iargs_S_DS.S=s0; Iargs_S_DS.DS=ds1}))
                                                                                                                //eprintfn "here1"
                                                                                                                LogInfo("here1")

                                                                                                                limitedmap (n+1) (i1' :: bag) instrs


                                                               | (XsXds(op1, {Iargs_S_DS.S=s1; Iargs_S_DS.DS=ds1}))::instrs
                                                                 when ds1 = d0
                                                                   && not(Set.contains (Operand_of_X  s1) out0)
                                                                   && not(Set.contains (Operand_of_X ds1) out0)
                                                                   && XsXds_is_commutative(op1) 
                                                                   && nextReader_of b (Operand_of_X s1) instrs = None
                                                                      -> if s1 = ds1 then None
                                                                                     else match try__M128sXds_of_XsXds(op1)
                                                                                           with | None -> None

                                                                                                | Some(op1') ->
                                                                                                    //eprintfn "here3"
                                                                                                    LogInfo("here3")

                                                                                                    let i1' = (M128sXds (op1',{Iargs_S_DS.S=s0; Iargs_S_DS.DS=s1}))
                                                                                                    let instrs' = List.map (fun(i) -> Map_X_in_Instr (swap2 s1 ds1) i) instrs
                                                                                                    Some(i1'::bag,instrs')


                                                               | (XsXd(op1, {Iargs_S_D.S=s1;Iargs_S_D.D=d1}))::instrs
//                                                               | (XsXd(((I_XsXd.Movaps | I_XsXd.Movapd) as op1), {Iargs_S_D.S=s1;Iargs_S_D.D=d1}))::instrs
                                                                 when s1 = d0
                                                                      -> if s1 = d1 then None
                                                                                    else match try__M128sXd_of_XsXd(op1)
                                                                                          with | None -> None
                                                                                               | Some(op1') -> let i1' = (M128sXd(op1',{Iargs_S_D.S=s0; Iargs_S_D.D=d1}))
                                                                                                               //eprintfn "here2"
                                                                                                               LogInfo("here2")
                                                                                                               limitedmap (n+1) (i1' :: bag) instrs
                                                                                                                                          
                                                                                                                                          
                                                               | i::instrs3 ->
                                                                   if   Set.contains (Operand_of_X d0) (D_of_Instr  i) then Some(bag,(i::instrs3))
                                                                   elif Set.contains (Operand_of_X d0) (DS_of_Instr i) then None
                                                                   elif Set.contains (Operand_of_X d0) (S_of_Instr  konstSizeDict b.ReturnsVoid i) then None
                                                                   elif (match access
                                                                          with | LocalRO |LocalRW(_) -> false
                                                                               | NonLocal -> not(Set.isEmpty(Set.intersect (Set.union (Set.ofList (m0 |> R32_of_M |> List.map (Rlong >> Operand_of_R)))
                                                                                                                                      (Set.ofList (m0 |> R64_of_M |> List.map (Rquad >> Operand_of_R))))
                                                                                                                           (Set.union (D_of_Instr(i)) (DS_of_Instr(i)))))
                                                                         )
                                                                        then
                                                                             LogInfo("OJE")
//                                                                             eprintfn "OJE"
                                                                             None 
//                                                                   elif (*!!!*)  nextReader_of b (Operand_of_X d0) (i::instrs3) = None then Some(bag,(i::instrs3))
                                                                   elif (match access
                                                                          with | LocalRO     -> false
                                                                               | LocalRW(m0) -> (Mds_of_Instr(i) @ Md_of_Instr(i)) |> List.choose stackLocOption |> List.exists (nonoverlapping_stackLoc m0 >> not)
                                                                               | NonLocal    -> (Mds_of_Instr(i) @ Md_of_Instr(i)) |> List.exists (fun m -> Raccess_of_M(m) = NonLocal))
                                                                        then None
(*
                                                                         (((Mds_of_Instr(i) @ Md_of_Instr(i))
                                                                                               |> List.exists (fun(m) -> stackLocOption(m) = None)) 
                                                                                           || ((Rds_of_Instr(i) @ Rd_of_Instr(i))
                                                                                               |> List.exists (fun(r:R) ->
                                                                                                                 R64_of_M(m0) // hoist
                                                                                                                 |> List.exists (fun(x:R64) -> r.Aliases(Rquad(x))))))) then None
*)                                                                                                                 
                                                                                                
                                                                   else
                                                                       LogInfo("here3")
//                                                                       eprintfn "here3 %A" (List.length instrs3)
                                                                       limitedmap n (i::bag) instrs3

                                match limitedmap 0 [] instrs with
                                    | Some(pre',post) ->
                                        loop bag             ((List.rev pre') @ post)
                                    | None            ->
                                        loop (i::bag) instrs
                                
                            | XsXd((I_XsXd.Movaps | I_XsXd.Movapd), {Iargs_S_D.S=s0; Iargs_S_D.D=d0}) ->
                                  if s0 = d0 then (* elim useless vector copy *) loop bag instrs
                                  elif Set.contains (Operand_of_X d0) out0 || Set.contains (Operand_of_X s0) out0 then loop (i::bag) instrs
                                  elif nextReader_of b (Operand_of_X d0) instrs = None then
                                      eprintfn "elim_useless_vec_copies_block: elim DC movaps"
                                      loop bag instrs
                                      
                                  else match nextReader_of b (Operand_of_X s0) instrs
                                        with | None -> // LogInfo("movaps/d movelim1")
                                                       loop bag (List.map (fun(i) -> (Map_X_in_Instr (swap2 s0 d0) i)) instrs)
                                             | Some(pre',(XsXds(op1,{Iargs_S_DS.S=s1; Iargs_S_DS.DS=ds1})),post) 
                                               when s1 <> ds1
                                                 && ds1 <> d0
                                                 && s1 = s0
                                                 && nextReader_of b (Operand_of_X s0) post = None
                                                 && pre' |> List.forall (fun(i) -> not(Set.contains (Operand_of_X ds1) (D_of_Instr  i) ||
                                                                                       Set.contains (Operand_of_X ds1) (DS_of_Instr i) ||
                                                                                       Set.contains (Operand_of_X ds1) (S_of_Instr  konstSizeDict b.ReturnsVoid i)))
                                                    -> //LogInfo(sprintf "movaps/d movelim2")
                                                       (List.rev pre') @ post
                                                       |> List.map (fun(i) -> (Map_X_in_Instr (swap2 s0 d0) i))
                                                       |> loop ((XsXds(op1,{Iargs_S_DS.S=s1; Iargs_S_DS.DS=ds1})) :: bag)

                                             | Some _ -> match nextReader_of b (Operand_of_X d0) instrs
                                                          with | Some(pre',(XsXds(op1,{Iargs_S_DS.S=s1; Iargs_S_DS.DS=sd1})),post)
                                                                 when sd1<>s1 && sd1=d0 && sd1<>s0 && s1<>s0
                                                                   && XsXds_is_commutative op1
                                                                   && nextReader_of b (Operand_of_X s1) post = None
                                                                   && not (List.exists (fun(i) -> Set.contains (Operand_of_X s0) (D_of_Instr i) || Set.contains (Operand_of_X s0) (DS_of_Instr i)) pre')
                                                                   && not (Set.contains (Operand_of_X s1) out0)
                                                                      -> //LogInfo("movopt2B")
                                                                         loop bag
                                                                              ((List.rev pre') @
                                                                               [(XsXds(op1, {Iargs_S_DS.S=s0;Iargs_S_DS.DS=s1}))] @
                                                                               (List.map (fun(i) -> (Map_X_in_Instr (swap2 s1 sd1) i)) post))
                                                               | _ -> match nextReader_of b (Operand_of_X s0) instrs
                                                                       with | Some(pre',(XsXds(op1,{Iargs_S_DS.S=s1; Iargs_S_DS.DS=sd1})),post)
                                                                              when sd1<>s1 && sd1=s0 && sd1<>d0 && s1<>s0 
                                                                                && XsXds_is_commutative op1
                                                                                && nextReader_of b (Operand_of_X s1) (pre' @ post) = None
                                                                                && not (List.exists (fun(i) -> Set.contains (Operand_of_X s1) (D_of_Instr i) || Set.contains (Operand_of_X s1) (DS_of_Instr i)) pre')
                                                                                && not (Set.contains (Operand_of_X s1) out0) 
                                                                                   -> //LogInfo("movopt2C")
                                                                                      loop ((XsXds(op1, {Iargs_S_DS.S=sd1; Iargs_S_DS.DS=s1}))::bag)
                                                                                           (List.map (fun(i) -> (Map_X_in_Instr (rot3 d0 sd1 s1) i)) ((List.rev pre') @ post))
                                                                            | _ -> loop (i::bag) instrs
                            | _ -> loop (i::bag) instrs

                b.Instrs
//                |> Array.map (fun(i) -> (single_instr_peephole_opt i))
                |> Array.toList
                |> moveDown_Instrs_like b (Instr_is_Vec >> not)
                |> moveDown_Instrs_like b (function XsXd((I_XsXd.Movapd | I_XsXd.Movaps),_) -> true | _ -> false)
                |> moveDown_Instrs_like b (function M128sXd((I_M128sXd.Movapd | I_M128sXd.Movaps), {Iargs_S_D.S=(M128.S(K(Id({Name = k})) | K64rip(Id({Name = k})))); Iargs_S_D.D=d}) -> konstSizeDict.ContainsKey(k) | _ -> false)
                |> loop []
                |> fun instrs1 -> b.Instrs <- Array.ofList(instrs1)


let elim_useless_vec_copies (tDict:Dictionary<String,IBlock>) (inoutDict:Dictionary<String,Set<Operand> * Set<Operand>>) =
    for KeyValue(name,b) in tDict do
      if inoutDict.ContainsKey(name) then
        let (in0,out0) = inoutDict.[name]
        elim_useless_vec_copies_block (in0,out0) b


let elim_nops (tDict:Dictionary<String,IBlock>) (cc:String []) =
    cc |> Array.iter (fun lbl -> let b = tDict.[lbl]
                                 b.Instrs <- b.Instrs |> Array.filter (Instr_is_nop >> not))
